/****************************************************************************
* File............: pkdir.c                                                 *
* Version.........: $Id: 2c4dd6185852650cc780ac5a7b568b7d3173f298 $         *
* Created.........: 2019-06-02                                              *
* Last changed....: 2019-08-18                                              *
* Author..........: TheBlackzone <theblackzone@gmail.com>                   *
*---------------------------------------------------------------------------*
* Purpose:                                                                  *
*   Unpack !PackDir archives                                                *
*---------------------------------------------------------------------------*
* Compile:                                                                  *
*   gcc -o pkdir pkdir.c                                                    *
*---------------------------------------------------------------------------*
* Usage:                                                                    *
*   pkdir {-d} [archive file]                                               *
*   -d = dry run, show what happens without actually doing it               *
****************************************************************************/


/*-------- Includes -------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>
#include <linux/limits.h>


/*-------- Defines --------------------------------------------------------*/

#define	RISCOS_NAME_MAX	(255)     /* max. lenght of file object name       */


/*-------- Constants ------------------------------------------------------*/

static const char *pgm_name = "pkdir";
static const char *pgm_purpose = "Extract Acorn !PackDir Archives";

static const unsigned long int ROOT_OFFSET = 9UL; /*    Offset of root dir */
static const int lzwbits = 8;        /* always use 8 bits GIF LZW algoritm */


/*-------- Object information ---------------------------------------------*/

typedef struct objectinfo_t
{
	char			name[RISCOS_NAME_MAX];
	unsigned int	load_addr;
	unsigned int	exec_addr;
	unsigned int	file_type;
	int64_t			timestamp;
	unsigned int	original_length;
	unsigned int	attributes;
	unsigned int	object_type;
	int				compressed_length;
} OBJECTINFO_T;


/*-------- Global vars ----------------------------------------------------*/

static int opt_dryrun;                             /* Flag for option '-d' */


/****************************************************************************
* info:                                                                     *
* Print info about this program.                                            *
****************************************************************************/

static void info(void)
{
	printf("%s - %s\n", pgm_name, pgm_purpose);
	printf("Usage: %s [packdirfile]\n", pgm_name);
} /* info */


/****************************************************************************
* getint32:                                                                 *
* Read and return a 32bit integer value from buffer at adress 'adr'.        *
* Integers are stored as 'little endian' in !PackDir files.                 *
* RETURNS the read integer.                                                 *
****************************************************************************/

static int getint32(const char *adr)
{
	char byte0,byte1,byte2,byte3;
	int int32;

	int32 = 0;

	/* 32-bit int values are stored as little endian */
	byte0 = adr[0];
	byte1 = adr[1];
	byte2 = adr[2];
	byte3 = adr[3];

	int32 = ((byte3 << 24) & 0xFF000000) | ((byte2 << 16) & 0x00FF0000) | ((byte1 << 8) & 0x0000FF00) | (byte0 & 0x000000FF);

	return(int32);

} /* getint32 */


/****************************************************************************
* get_objectinfo:                                                           *
* Reads the information on a stored object from the buffer 'data' starting  *
* at offset 'offset' and puts the retrieved information into the struct     *
* 'objinfo'.                                                                *
* RETURNS the new offset after reading the data                             *
****************************************************************************/

static unsigned long int get_objectinfo(const char *data, unsigned long int offset, 
		struct objectinfo_t *objinfo)
{
	unsigned long int i, j;
	int isroot;
	char name[RISCOS_NAME_MAX];

	memset(objinfo, 0, sizeof(struct objectinfo_t));

	/* Set flag, if we are at the root position */
	isroot = 0;
	if (offset == ROOT_OFFSET)
		isroot = 1;
	
	i = offset;
	
	/* Get object name. 
	 * RISC OS allows '/' in names which are commonly used to 
	 * seperate the 'file extension'. We replace them with a '.' */
	j = 0;
	do {
		name[j] = (data[i] == '/') ? '.' : data[i];
		j++;
		i++;
	} while ((data[i - 1] != '\0') && (j < RISCOS_NAME_MAX));

	if (isroot == 0) {
		strncpy((char*)&objinfo->name, name, RISCOS_NAME_MAX);
	} else {
		/* Get only basename if we have the root dir entry */
		j = (unsigned long int)strlen(name);
		while (j > 0 && name[j - 1] != '.')
			j--;
		strncpy((char*)&objinfo->name, &name[j], RISCOS_NAME_MAX);
	}

	/* Get object load address and exec address. These fields hold
	 * the file type and timestamp */
	objinfo->load_addr = (unsigned int)getint32(data + i);
	i += 4;
	objinfo->exec_addr = (unsigned int)getint32(data + i);
	i += 4;
	
	/* Filetype */
	objinfo->file_type = 0;
	if ((objinfo->load_addr & 0xFFF00000) == 0xFFF00000)
		objinfo->file_type = (objinfo->load_addr >> 8) & 0x00000FFF;

	/* Timestamp. RISC OS uses timestamps in centiseconds since 1900-01-01
	 * but Unix uses seconds since 1970-01-01. We need to do some 
	 * conversion here. */
	objinfo->timestamp = (int64_t)(objinfo->load_addr & 0x000000FF);
	objinfo->timestamp = (int64_t)((objinfo->timestamp << 32) | objinfo->exec_addr);
	objinfo->timestamp = (int64_t)(objinfo->timestamp - 0x336E996A00); /* 0x336E996A00 is the amount of centiseconds from 1900-01-01 (RISC OS epoch) till 1970-01-01 (Unix epoch) */
	objinfo->timestamp = (int64_t)(objinfo->timestamp / 100); /* convert to seconds */

	/* Original data length (if file) or number of entries (if directory) */
	objinfo->original_length = (unsigned int)getint32(data + i);
	i += 4;

	/* File attributes */
	objinfo->attributes = (unsigned int)getint32(data + i);
	i += 4;

	/* At the root entry, the object type is omitted because it is always a
	 * directory. In this case we set it manually. Else we read the object
	 * type field */
	if (isroot == 1) {
		objinfo->object_type = 1;
	} else {
		objinfo->object_type = (unsigned int)getint32(data + i);
		i += 4;
	}

	/* If the object is a file, the length of the compressed data follows
	 * (this field is omitteded if the object is a directory. If the length
	 * is (-1) then the data is uncompressed and just copied from the buffer,
	 * if the length is (-2) no data is stored at all */
	objinfo->compressed_length = 0;
	if (objinfo->object_type == 0)
	{
		objinfo->compressed_length = getint32(data + i);
		i += 4;
	}

	/* return the index we are currently at */
	return(i);
} /* get_objectinfo */


/****************************************************************************
* mem2file:                                                                 *
* Writes a block of data of length 'len' from the buffer at 'data' to a     *
* file with the name 'filename'.                                            *
* RETURNS 0 on error, 1 on success                                          *
****************************************************************************/

static unsigned int mem2file(void *data, unsigned int len, 
		const char *filename)
{
	FILE *f;
	unsigned int i;

	i = 0;

	f = fopen(filename, "wb");
	if (!f) return(i);

	i = (unsigned int)fwrite(data, (size_t)len, (size_t)1, f);
	fclose(f);
	return(i);
} /* mem2file */


/****************************************************************************
* decompress:                                                               *
* Decompress a block of data at 'data' with a compressed length of 'clen'   *
* and expand it to the block at 'outbuf' with a maximum lenght of 'olen'    *
* using the GIF LZW algorithm with 'bits' bits of code length               *
****************************************************************************/

static void decompress(unsigned char *data, int clen, 
		unsigned char *outbuf, unsigned long int olen, int bits)
{
	unsigned long int max;
	unsigned long int i;
	unsigned long *prefix;
	unsigned char *suffix;
	unsigned char *stack;
	unsigned char *stackptr;
	unsigned long int codesize;
	unsigned long int codemask;
	unsigned long int cc;
	unsigned long int eoi;
	unsigned long int nextfree;
	unsigned long int code;
	unsigned long int prev_code;
	unsigned long int save_code;
	unsigned long int temp_code;
	unsigned char firstchar;
	unsigned long int bitbuffer;
	unsigned long int bitsremain;
	unsigned long int outptr;

	/* The maximum number of codes is derived from the bits used.
	 * For example: 13 bits results in 1<<13 = 8192 codes */
	max = (unsigned long int)(1 << bits);

	/* Allocate memory for code tables */
	prefix = (unsigned long*)malloc(sizeof(unsigned long) * (max + 1));
	suffix = (unsigned char*)malloc(sizeof(unsigned char) * (max + 1));
	stack  = (unsigned char*)malloc(sizeof(unsigned char) * (max + 1));

	if (prefix == NULL || suffix == NULL || stack == NULL) {
		printf("Insufficient memory. Abort.\n");
		exit(EXIT_FAILURE);
	}

	codesize = (unsigned long int)lzwbits + 1;
	codemask = (unsigned long int)((1 << codesize) - 1);
	cc = (unsigned long int)(1 << lzwbits);
	eoi = (cc + 1);
	nextfree = cc + 2;

	/* Initialize tables */
	for (i = 0; i < cc; i++) {
		prefix[i] = 0L;
		suffix[i] = (unsigned char)i;
		stack[i] = 0;
	}

	/*
	printf("Codesize.........: %u\n", codesize);
	printf("Codemask.........: %04x\n", codemask);
	printf("Clear Code.......: %u\n", cc);
	printf("EOI..............: %u\n", eoi);
	printf("Nextfree.........: %u\n", nextfree);
	*/

	i = 0;
	prev_code = 0xFFFFFFFF;
	bitbuffer = 0L;
	bitsremain = 0L;
	stackptr = &stack[0];
	outptr = 0;

	while (i < (unsigned long int)clen) {
		bitbuffer = bitbuffer | (data[i] << bitsremain);
		bitsremain += lzwbits;
		i++;

		/* Decode while there are enough bits in the buffer */
		while (bitsremain > codesize) {
			code = bitbuffer & codemask;
			bitsremain -= codesize;
			bitbuffer = bitbuffer >> codesize;

			/* Is it a CLEAR code? Reinitialize. */
			if (code == cc) {
				codesize = (unsigned long int)(lzwbits + 1);
				codemask = (1 << codesize) - 1;
				cc = (unsigned long int)(1 << lzwbits);
				eoi = (cc + 1);
				nextfree = cc + 2;
				prev_code = 0xFFFFFFFF;
				continue;
			}

			/* Handle EOI code */
			if (code == eoi) {
				i = (unsigned long int)clen; /* Force end of main loop */
				break;
			}

			/* First code of the stream? Direct output */
			if (prev_code == 0xFFFFFFFF) {
				if (outptr < olen)
					outbuf[outptr++] = suffix[code];
				prev_code = code;
				firstchar = (unsigned char)code;
				continue;
			}

			/* -- Decoding part --- */

			save_code = code;
			if (code >= nextfree) {
				*stackptr++ = firstchar;
				code = prev_code;
			}

			while (code > cc) {
				temp_code = code;
				if (code == prefix[code]) {
					fprintf(stderr, "Circular data error! (1)");
					exit(EXIT_FAILURE);
				}

				*stackptr++ = suffix[code];
				code = prefix[code];

				if (temp_code == prefix[code]) {
					fprintf(stderr, "Circular data error! (2)");
					exit(EXIT_FAILURE);
				}
			}

			firstchar = suffix[code];
			*stackptr++ = suffix[code];

			if (nextfree < max) {
				prefix[nextfree] = prev_code;
				suffix[nextfree] = firstchar;
				nextfree++;

				/* Increase the code size if all codes for the current
				 * bit length are used */
				if (((nextfree & codemask) == 0) && (nextfree < max)) {
					codesize++;
					codemask += nextfree;
				}

			}
			prev_code = save_code;

			/* Output what has been accumulated on the stack */
			while (stackptr > stack) {
				stackptr--;
				if (outptr < olen)
					outbuf[outptr++] = *stackptr;
			} 

		}
	}

	free(prefix);
	free(suffix);
	free(stack);

} /* decompress */


/****************************************************************************
* stamp:                                                                    *
* Set the modifcation timestamp of a file/directory 'name' to 'timestamp'   *
****************************************************************************/

static void stamp(const char *name, const int64_t timestamp)
{
	int rc;
	struct stat st;
	struct utimbuf newtime;

	rc = stat(name, &st);
	if (rc == -1) return;

	newtime.actime = st.st_atime;
	newtime.modtime = (__time_t)timestamp;
	utime(name, &newtime);
} /* stamp */



/****************************************************************************
* create_dir:                                                               *
* Create a directory 'name' with timestamp 'timestamp'                      *
* RETURNS 0 on success, -1 on failure                                       *
****************************************************************************/

static int create_dir(const char *name, int64_t timestamp)
{
	struct stat st;
	int rc;

	rc = 0;
	if (stat(name, &st) == -1)
		rc = mkdir(name, 0777);
	if (rc == 0)
		stamp(name, timestamp);
	return(rc);
} /* create_dir */



/****************************************************************************
* read_dir:                                                                 *
* Recursively read through directories in the buffer 'data' starting at     *
* offset 'offset'. 'num_entries' are the number of items in a directory,    *
* 'dir' holds the directory name used to build object names recursively.    *
* 'bits' are the number of LZW compression bits they are just passed on to  *
* the decompress function                                                   *
* 'depth' is the recursion depth for indented display of the object names.  *
* RETURNS the new offset in the data after the directory has been worked    *
* on.
****************************************************************************/

static unsigned long int read_dir(char *data, unsigned long int offset, 
		unsigned int num_entries, unsigned int depth, const char *dir, int bits)
{
	unsigned long int i;
	struct objectinfo_t oi;
	unsigned int e,j;
	char dirname[PATH_MAX];
	char objname[PATH_MAX];
	void *outbuf;
	int rc;
	
	rc = -1;
	/* If we are at the root offset (ie. first entry)  we need to handle the 
	 * root directory object first */
	if (offset == ROOT_OFFSET) {
		i = get_objectinfo(data, offset, &oi);
		e = oi.original_length;
		if (opt_dryrun != 1)
			rc = create_dir(oi.name, oi.timestamp);
		strncpy(dirname, oi.name, PATH_MAX-1);
		if (opt_dryrun == 1)
			printf("DRY RUN : ");
		printf("ROOT DIR: %s\n", dirname);
	} else {
		i = offset;
		e = num_entries;
		if (opt_dryrun != 1)
			rc = create_dir(dir, time(NULL));
		strncpy(dirname, dir, PATH_MAX);
	}

	/* Directory creation failed? */
	if (rc == -1) {
		fprintf(stderr, "Cannot create directory '%s'. Abort.", oi.name);
		exit(EXIT_FAILURE); /* TODO: Better error handling */
	}

	while (e > 0) {
		i = get_objectinfo(data, i, &oi);
		if (oi.object_type == 0) { /* FILE */
			rc = snprintf(objname, PATH_MAX-1, "%s/%s,%03x", dirname, oi.name, oi.file_type);
			if (rc < 0) {
				fprintf(stderr, "Output filename and path are too long. Skipping.\n");
			} else {
				for (j=0; j < depth * 2; j++)
					printf(" ");
				if (opt_dryrun == 1)
					printf("DRY RUN : ");
				printf("UNPACK FILE: %s\n", objname);
				if (oi.compressed_length == -1) {
					if (opt_dryrun != 1)
						mem2file(&data[i], oi.original_length, objname);
					i += oi.original_length;
				} else {
					if (opt_dryrun != 1) {
						outbuf = malloc((size_t)oi.original_length);
						if (outbuf) {
							decompress((unsigned char*)&data[i], oi.compressed_length, outbuf, oi.original_length, bits);
							mem2file(outbuf, oi.original_length, objname);
							free(outbuf);
						}
					}
					i += oi.compressed_length;
				}
				if (opt_dryrun != 1)
					stamp(objname, oi.timestamp);
			}
		} else { /* DIRECTORY */
			for (j=0; j < depth * 2; j++)
				printf(" ");
			if (opt_dryrun == 1)
				printf("DRY RUN : ");
			rc = snprintf(objname, PATH_MAX-1, "%s/%s", dirname, oi.name);
			if (rc < 0) {
				fprintf(stderr, "Output filename and path are too long. Skipping.\n");
			} else {
				printf("UNPACK DIR: %s\n", objname);
				i = read_dir(data, i, oi.original_length, depth + 1, objname, bits);
				if (opt_dryrun != 1)
					stamp(objname,oi.timestamp);
			}
		}
		e--;
	}
	return(i);

} /* read_dir */


int main(int argc, char **argv)
{
	FILE *f;
	int rc;
	unsigned long int filelen;
	char *data;
	int bits;
	int i;
	unsigned long int offset;
	char filename[FILENAME_MAX];

	if (argc < 1) {
		info();
		exit(EXIT_FAILURE);
	}

	/* init */
	memset(filename, 0, (size_t)FILENAME_MAX);
	opt_dryrun = 0;

	/* get args */
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-d") == 0) {
			opt_dryrun = 1;
		} else if (argv[i][0] != '-') {
			strncpy((char*)&filename, argv[i], (size_t)FILENAME_MAX);
		} else {
			printf("Unknown option '%s'\n", argv[i]);
			exit(EXIT_FAILURE);
		}
	}

	if (strlen(filename) == 0) {
		printf("No filename specified.\n");
		exit(EXIT_FAILURE);
	}

	f = fopen(filename, "rb");
	if (f == NULL) {
		fprintf(stderr, "Cannot open file '%s'. Abort.\n", filename);
		exit(EXIT_FAILURE);
	}

	rc = fseek(f, 0L, SEEK_END);
	if (rc != 0) {
		perror("Cannot determine file length");
		fclose(f);
		exit(EXIT_FAILURE);
	}

	filelen = (unsigned long int)ftell(f);

	printf("File length is %lu bytes.\n", filelen);

	data = (char*)malloc((size_t)filelen);
	if (data == NULL) {
		perror("Cannot alocate memory to read file content");
		fclose(f);
		exit(EXIT_FAILURE);
	}

	rewind(f);
	rc = (int)fread(data, (size_t)filelen, (size_t)1, f);

	fclose(f);

	if (rc != 1) {
		printf("Return %d\n",rc);
		perror("File read is incomplete");
		free(data);
		exit(EXIT_FAILURE);
	}

	/* The first four bytes of a !PackDir file read "PACK" */
	rc = strncmp((char*)data, "PACK", (size_t)4);
	if (rc != 0) {
		fprintf(stderr, "File is not a !PackDir archive! Abort.\n");
		free(data);
		exit(EXIT_FAILURE);
	}

	bits = getint32(data + 5) + 12;
	printf("This !PackDir file uses %d-bit LZW compression.\n", bits);

	offset = ROOT_OFFSET;
	while (offset < filelen - 1) {
		offset = read_dir(data, offset, 0, 0, "", bits);
	}
		
	free(data);

	return(EXIT_SUCCESS);

} /* main*/

